// File:         $Id: Game.h,v 1.20 2012/11/08 01:49:02 jbf9160 Exp $
// Author:       James Fator (jbf9160)
// Author:       Ben Rajcan (bmr2238)
// Description:  Game is the inherited class for all the two person
//                   games in the CS4 Project.
// Revisions:    
//               $Log: Game.h,v $
//               Revision 1.20  2012/11/08 01:49:02  jbf9160
//               Minor updates to style.
//
//               Revision 1.19  2012/10/29 10:35:33  jbf9160
//               Improved solution algorithm.
//
//               Revision 1.18  2012/10/27 02:15:28  jbf9160
//               Updated solution iterator to choose definite best move.
//
//               Revision 1.17  2012/10/22 00:45:58  jbf9160
//               Concurrent implementation.
//                   Solution for Computer Player will now
//                   be created on a separate thread to allow
//                   for quicker return time.
//
//               Updated isGoodStateToBeIn method to return bool.
//
//               Revision 1.16  2012/10/15 01:05:49  jbf9160
//               Small tweaks.
//               Updated commenting standard.
//
//               Revision 1.15  2012/10/14 08:58:51  bmr2238
//               Game framework updated to support Kayles.
//
//               Revision 1.14  2012/10/11 03:16:14  jbf9160
//               Game framework completed. Takeaway Game finished.
//
//               Revision 1.13  2012/10/10 10:25:00  jbf9160
//               Computer class working.
//
//               Revision 1.12  2012/10/10 07:59:48  jbf9160
//               Game is now an abstract class that includes all
//                   the necessary methods to play the game.
//               Game now includes the Players directly which makes
//                   it easy to implement them.
//               Human class is up and running but the Computer
//                   class requires testing.
//
//               Revision 1.11  2012/10/09 05:19:17  jbf9160
//               Game file working as framework now.
//
//               Revision 1.10  2012/10/09 02:27:21  jbf9160
//               Combined game with the framework.
//
//               Revision 1.9  2012/10/08 03:48:24  jbf9160
//               Commit to initiate rough scale testing.
//
//               Revision 1.8  2012/10/08 00:39:00  jbf9160
//               Game interface completed... for now...
//
//               Revision 1.7  2012/10/04 03:17:41  jbf9160
//               Updated for solution algorithm.
//
//               Revision 1.6  2012/10/03 02:12:42  jbf9160
//               Modified the framework to accept Game for a
//                  Player or computer.
//
//               Revision 1.5  2012/10/03 00:04:38  jbf9160
//               Game class will now support dynamic variables
//                  in tpgframework.
//
//               Revision 1.4  2012/10/02 05:26:28  jbf9160
//               Minor bug fix.
//
//               Revision 1.3  2012/10/02 05:18:55  jbf9160
//               Implemented strategy method.
//
//               Revision 1.2  2012/10/02 04:59:20  jbf9160
//               Game class fully created with all required elements to
//                   be accessed in tpgframework.
//
//               Revision 1.1  2012/10/02 04:04:21  jbf9160
//               Initial commit of the Game interface.
//               Skeleton created.
//

#ifndef GAME_H_
#define GAME_H_

#include <iostream>
#include <cstdlib>
#include <vector>
#include <map>
#include <pthread.h>

/*
 * Game is the interface for each game in the CS4 Project
 *     Game can be implemented through one of the game source files.
 *     Game has a template of < class for state, class for choices >
 *
 * @precondition to implementing the Game class:
 *     if the STATE contains const variables, then it must implement
 *         operator= method.
 *     the STATE data type must implement operator< method.
 *     the CHOICE data type must implement a default constructor.
 *     the CHOICE data type must implement operator< method.
 */
template<typename STATE, typename CHOICE>
class Game {

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
public: /* Game : Constructor && Destructor */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
    /*
     * Constructor.
     * If at all possible, we would prefer using the newGame method.
     *
     * @precondition: state is a valid STATE in the Game
     *
     * @param state - the initial STATE of the Game.
     */
    Game(const STATE& state) :
            _state( new STATE( state ) ) {
        // Nothing more to construct...
    } /* Game : Constructor */

    /*
     * Destructor to make the compiler happy...
     */
    virtual ~Game() {

        // If there is an allocated STATE,
        if ( _state ) {

            // DESTROY IT!
            delete _state;
        }
    } /* Game : Destructor */

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
public: /* Game : Pure virtual methods */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

    /*
     * getState will return the STATE of the Game based on the
     *     theoretical CHOICE made
     *
     * @precondition: state is a valid STATE in the Game
     * @precondition: choice is a valid CHOICE in the Game
     * @postcondition: returned STATE is a valid STATE in the Game
     *
     * @param state - the STATE at which to determine the
     *                output of a provided CHOICE
     * @param choice - the CHOICE to make at the provided STATE
     * @return the STATE at which the Game moves to after
     *         the provided CHOICE
     */
    virtual STATE getState(const STATE& state, const CHOICE& choice)
                                                            const = 0;

    /*
     * getChoices returns the choices that the player can pick from
     *     at a specific point in the game.
     *
     * @precondition: state is a valid STATE in the Game
     * @postcondition: vector returned contains all valid choices
     *                 for that STATE
     *
     * @param state - the state to get choices from.
     * @return the vector of CHOICEs.
     */
    virtual std::vector<CHOICE> getChoices(const STATE& state)
                                                            const = 0;
    /*
     * isFinalState tests if the provided STATE is an accepting STATE.
     *
     * @precondition: state is a valid STATE in the Game
     *
     * @param state - the STATE to test if accepting
     * @return true if the provided STATE is an accepting STATE
     */
    virtual bool isFinalState(const STATE& state) const = 0;

    /*
     * isGoodStateToBeIn determines if a state that you leave your opponent
     *     with will leave him at a disadvantage.
     *
     * @precondition: state is a valid STATE in the Game
     *
     * @param state - the STATE to test if good to be at
     * @return 0 this is not a good STATE.
     *         1 or -1 if this is a good STATE.
     *         In the event of a final STATE: 0 if no winner,
     *         1 if player 1 won, -1 if player two won.
     */
    virtual int isGoodStateToBeIn(const STATE& state) const = 0;

    /*
     * stringToChoice takes a string input and converts it to a CHOICE.
     *
     * stringToChoice will also check the validity of the CHOICE
     *
     * @postcondition: If returned true, then choice argument is modified
     *                 to a valid CHOICE for the Game.
     *
     * @param input - the string inputed to convert to CHOICE
     * @param choice - the CHOICE holder to modify if valid input
     * @return true if the input was valid. false if not.
     */
    virtual bool stringToChoice(std::string& input, CHOICE& choice) const = 0;

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
public: /* Game : Output methods */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

    /*
     * printState will print the current STATE of the game for the
     *     human counterpart. This is pure virtual because you should
     *     add context to the STATE when outputting.
     *
     * @precondition: stream argument starts where it wants to start
     *                outputting the STATE.
     *
     * @postcondition: output ends immediately right after STATE is
     *                 outputed.
     *
     * @param out - the output stream to print to.
     */
    virtual void printState(std::ostream& out) const = 0;

    /*
     * printChoices is how the Player player will know what the available
     *     CHOICEs are. Prints to Standard Out.
     *
     * printChoices does not have to print all the CHOICEs out if there are
     *     too many; it may print out an overview of what the next choice
     *     should be.
     *
     * @precondition: stream argument starts where it wants to start
     *                outputting the CHOICEs.
     *
     * @postcondition: output ends immediately right after CHOICEs are
     *                 outputed.
     *
     * @param out - the output stream to print to.
     */
    virtual void printChoices(std::ostream& out) const = 0;

    /*
     * printAChoice is how the Computer Player
     *     prints out the CHOICE it made.
     *
     * @precondition: stream argument starts where it wants to start
     *                outputting the CHOICE.
     *
     * @postcondition: output ends immediately right after CHOICE is
     *                 outputed.
     *
     * @param choice - the CHOICE to be printed
     * @param out - the output stream to print to.
     */
    virtual void printAChoice(const CHOICE& choice,
            std::ostream& out) const = 0;

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
public: /* Game : Non virtual methods */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
    /*
     * currentState will return the current STATE of the game
     *
     * @return the current STATE of the game
     */
    const STATE& currentState() const {

        // Return the STATE of the Game
        return *_state;

    } /* Game : currentState */

    /*
     * play is the main class to actually play the game with.
     * play moves back and forth between players until a final
     *     STATE is reached.
     *
     * @param playerOne - if true, then human player; false will
     *                        create computer instance
     * @param playerTwo - if true, then human player; false will
     *                        create computer instance
     */
    void play(bool playerOne, bool playerTwo) {

        // Create backup STATE for multiple plays:
        const STATE initialState = currentState();

        // Create temporary players
        Player* _playerOne = 0;
        Player* _playerTwo = 0;

        // Create Player One
        if (playerOne) {
            // Human Player!
            _playerOne = new Player(*this, 1);
        } else {
            // Computer Player!
            _playerOne = new Computer(*this, 1);
        }

        // Create Player Two
        if (playerTwo) {
            // Human Player!
            _playerTwo = new Player(*this, 2);
        } else {
            // Computer Player!
            _playerTwo = new Computer(*this, 2);
        }

        // Create current Player system
        Player* currentPlayer = _playerOne;

        // Initiate the primary loop!
        // This loop will continue until a re-match is
        // declined...
        while (true) {

            // Reset the output screen.
            std::cout << "\033[2J" << std::flush;
            std::cout << "\033[" <<  1 << ';'  << 1 << 'H' <<
                    std::flush;

            // Loop until the final state is reached...
            while (true) {

                // Clear the input.
//                std::cin.ignore();
//                std::cin.clear();

                // Set current STATE to a variable for easy access.
                STATE current = currentState();

                // Print the current STATE
                printState( std::cout );
                std::cout << std::endl;

                // Check if Final STATE
                if ( isFinalState(current) ) {
                    // Check to see if it was good to end up on
                    //     final state...
                    int winner( isGoodStateToBeIn( current ) );
                    if ( winner == 0 ) {

                        // There was a draw, no one gets points.
                        // Let the user know:
                        std::cout << "\nGame ended in a draw!" << std::endl;

                    } else if ( winner == 1 ) {

                        // The currentPlayer lost!

                        // Let the user know who won:
                        if (currentPlayer == _playerOne) {
                            _playerTwo->winForPlayer();
                            std::cout << "\nPlayer 2 wins!" << std::endl;

                        } else {
                            _playerOne->winForPlayer();
                            std::cout << "\nPlayer 1 wins!" << std::endl;
                        }

                    } else if ( winner == -1 ) {

                        // The current player wins!
                        // Increment the Player's score:
                        currentPlayer->winForPlayer();

                        // Let the user know who won:
                        if (currentPlayer == _playerOne) {
                            std::cout << "\nPlayer 1 wins!" << std::endl;
                        } else {
                            std::cout << "\nPlayer 2 wins!" << std::endl;
                        }
                    }
                    // Quit Game loop.
                    break;
                }

                // Get move from currentPlayer
                std::cout << "\nPlayer " << currentPlayer->getPlayerNumber()
                        << " - your turn." << std::endl;
                CHOICE move = currentPlayer->getMove(current);
                std::cout << std::endl << std::endl;

                // Update the Game STATE
                updateState(move);

                // Reset the output screen.
                std::cout << "\033[2J" << std::flush;
                std::cout << "\033[" <<  1 << ';'  << 1 << 'H' <<
                        std::flush;

                // Print the CHOICE made...
                std::cout << "Player " <<
                        currentPlayer->getPlayerNumber() << " chose: ";
                printAChoice( move, std::cout );
                std::cout << std::endl << std::endl;

                // Change the player!
                if (currentPlayer == _playerOne) {
                    currentPlayer = _playerTwo;
                } else {
                    currentPlayer = _playerOne;
                }
            }

            // Loop was broken which means someone won!
            // Print out Scores:
            std::cout << std::endl << "Current Scores:" << std::endl;
            std::cout << "\tPlayer 1: " << _playerOne->getScore()
                    << std::endl;
            std::cout << "\tPlayer 2: " << _playerTwo->getScore()
                    << std::endl;
            std::cout << std::endl;

            // Loop user input until we get what we want!
            while (true) {
                // Prompt user if they want to play again:
                std::cout << "Play again? ( y / n ): ";
                char again;
                std::cin >> again;

                // Clear the input.
                std::cin.ignore();
                std::cin.clear();

                // Check validity of input
                if (again == 'n') {

                    // Reset the output screen.
                    std::cout << "\033[2J" << std::flush;
                    std::cout << "\033[" <<  1 << ';'  << 1 << 'H';

                    // Print out Scores:
                    std::cout << "Final Scores:" << std::endl;
                    std::cout << "\tPlayer 1: " << _playerOne->getScore()
                            << std::endl;
                    std::cout << "\tPlayer 2: " << _playerTwo->getScore()
                            << std::endl;
                    std::cout << std::endl;
                    // Exit method.
                    return;

                } else if (again == 'y') {
                    // Reset the STATE to the initial state and loop around!
                    setState( initialState );
                    currentPlayer = _playerOne;
                    break;
                } else {
                    // They entered an incorrect character...
                    std::cout << "Incorrect character... try again!\n";
                }
            }
        }

        // Delete the Player pointers:
        delete _playerOne;
        delete _playerTwo;
    } /* Game : play */

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
private: /* Game : Player classes */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
    /*
     * Player is a Player class. The Player class will use
     *     standard output and input to receive Game choices
     */
    class Player {

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    public: /* Player : Constructor && Destructor */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
        /*
         * Player constructor creates an instance of a Player
         *
         * @param game the Game that will be played
         * @param playerNumber (1 OR 2) is the turn the Player will take.
         */
        Player(const Game& game, int playerNumber) :
                _game(&game), _number(playerNumber), _score( 0 ) {
            // Nothing to do here...
        } /* Player : Constructor */

        /*
         * ~Player is the destructor to make the compiler happy...
         */
        virtual ~Player() {
        }

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    public: /* Player : Accessor methods */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        /*
         * getPlayerNumber will return 1 or 2 depending on which player
         *      they are
         *
         * @return 1 or 2
         */
        int getPlayerNumber() const {
            return _number;
        } /* Player : getPlayerNumber */

        /*
         * getMove is inherited by all implemented Player classes
         *     It will take in the current state and return the desired move.
         *
         * @param state is the state of the game
         * @return a CHOICE of how to move.
         */
        virtual CHOICE getMove(const STATE& state) {
            // Create CHOICE
            CHOICE choice;

            // Loop until valid input
            while (true) {
                // Prompt user for number of pennies
                std::cout << "Enter move (type HELP for available choices): ";

                // Get input from user
                std::string input;
                // Get the entire line up until '\n'
                std::getline(std::cin, input, '\n');

                // Check to see if the user asked for help:
                if (input == "HELP" || input == "help" || input == "Help") {
                    // If so, print choices.
                    std::cout << "\033[2J" << std::flush;
                    std::cout << "\033[" <<  1 << ';'  << 1 << 'H';
                    _game->printState(std::cout);
                    std::cout << std::endl << std::endl;
                    _game->printChoices(std::cout);
                    std::cout << std::endl << std::endl;
                } else {
                    // Check to see if input was valid...
                    if (_game->stringToChoice(input, choice)) {
                        // if so, break the loop so we can return the choice.
                        break;
                    } else {
                        // Else, prompt, then continue the loop...
                        std::cout << "Invalid choice..." << std::endl;
                    }
                }
            }

            // Return the input...
            return choice;

        } /* Player : getMove */

        /*
         * getScore will return how many times the Player has won.
         *
         * @return the Player's score.
         */
        int getScore() const {

            return _score;

        } /* Player : getScore */

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    public: /* Player : Modifier methods */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        /*
         * winForPlayer will increment the Player's score.
         */
        void winForPlayer() {

            // Increment the score!
            _score++;

        } /* Player : winForPlayer */

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    protected: /* Player : Instance Variables */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        const Game* _game; // Game being played

        const int _number; // Player number

        int _score; // Score for the player!

    }; /* Player : class */

    /*
     * Computer is a Player class. The Computer class will create a
     *     solution based on the rules of the game.
     *
     * Computer uses a separate thread to construct the solution.
     *     The getMove method simply waits until the best CHOICE is found
     *     then returns that CHOICE. This was implemented because the program
     *     took a considerable amount of time to construct the solution
     *     while the rest of the program waited. This implementation will
     *     allow the entire program to run while the solution is being
     *     created.
     */
    class Computer: public Player {

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    public: /* Computer : Constructor && Destructor */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
        /*
         * Computer constructor kicks off the construction of a solution
         *     using concurrent programming.
         *
         * @param game the Game that will be played.
         * @param playerNumber (1 OR 2) is the turn the Player will take.
         */
        Computer(const Game& game, int playerNumber) :
                Player(game, playerNumber) {

            // Create a new pthread to construct the solution.
            pthread_t solutionConstructor;

            // Create the thread and start constructing!
            pthread_create(&solutionConstructor, NULL, threadHelper, this);

        } /* Computer : Constructor */

        /*
         * ~Solver is how the solution map is properly destroyed...
         */
        ~Computer() {

            // Clear the cache to conserve space...
            _cache.clear();

            // Delete the solution
            _solution.clear();

        } /* Computer : Destructor */

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    private: /* Computer : Public methods */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        /*
         * getMove is inherited by all implemented Player classes.
         *     It will take in the current state and return the desired move.
         *     getMove DOES NOT CHECK THE VALIDITY OF THE MOVE!!!
         *
         * @param state is the state of the game.
         * @return a CHOICE of how to move.
         */
        CHOICE getMove(const STATE& state) {

            std::cout << "Computer is thinking...\r" << std::flush;

            // Wait here until you have a pointer to
            //     a CHOICE
            while(! (_solution[state])) { }

            // Simply return the CHOICE mapped to that STATE.
            //    If we did it correctly, it will be the best choice.
            const CHOICE* move( _solution[state] );

            // Create temporary variable for the return CHOICE
            CHOICE choice = CHOICE();

            // Check to see if the CHOICE returned is the default CHOICE
            if ( !(move[0] < choice)  && !(choice < move[0]) ) {

                // Else, if it is... Let the players know there was no
                //     best CHOICE to make.
                std::cout << "No best choice...      ";

                // Set the choice to the first CHOICE available
                choice = this->_game->getChoices(
                                this->_game->currentState())[0];

                return choice;

            } else {

                // In the case it is not, let the players know...
                std::cout << "                       ";
                // Set the best CHOICE found to the CHOICE being returned.
                return CHOICE(move[0]);
            }
        } /* Computer : getMove */

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    private: /* Computer : Private methods */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        /*
         * threadHelper is a static method that kicks off the initial
         *      creation of the computer solution so that the rest
         *      of the program can run while the solution is created.
         *
         * @precondition: context MUST be a pointer to this Computer class.
         * @postcondition: getChoice will wait until a choice is found
         *                 to select a move.
         *
         * @param context - a pointer to the Computer class we are making a
         *                  solution for
         * @return null pointer to end the thread.
         */
        static void* threadHelper( void* context ) {

            // Get the context from the thread creation.
            Computer* computer = (Computer*) context;

            // Create a copy of the starting STATE
            STATE startingState( computer->_game->currentState() );

            // Initiate the solution creation!
            // (computer->_number % 2) will equal true if you are player one.
            computer->iterateSolution( startingState, computer->_number % 2 );

            // End the thread.
            pthread_exit(0);
        } /* Computer : threadHelper */

        /*
         * iterateSolution will move through every solution to determine
         *     which ones are the best to choose.
         *
         * @param state the STATE you're currently looking at
         * @param player true if it is your turn to choose on that state.
         * @return true if the path will eventually get to a good path.
         */
        int iterateSolution(const STATE& state, bool player) {

            // Check for final STATE:
            if ( this->_game->isFinalState( state ) ) {

                // Get whether or not the final STATE is good to be at:
                int score( this->_game->isGoodStateToBeIn( state ) );

                // Whatever score is, it will be the inverse if
                //     the Player is the opponent.
                if ( ! player ) {
                    score = score * -1;
                }

                // Set the CHOICE in the solution to NULL because
                //     we will never need to make a CHOICE at this STATE.
                _solution[state] = NULL;

                // To avoid the final state interfering with our numbers,
                //    we return zero.
                return score;
            }

            // Else, since it is not final: ALGORITHM!

            // Get available choices
            std::vector<CHOICE> choices( this->_game->getChoices( state ) );

            // If it is our turn...
            if (player) {

                // Create temporary for all next STATEs
                std::vector<STATE> nextStates;

                // Iterate through the choices set to get a
                //     vector of next STATEs
                for (unsigned int i = 0; i < choices.size(); i++) {

                    // Get the STATE that comes from the CHOICE made
                    STATE nextInLine = this->_game->getState(
                            state, choices[i]);

                    // Test if that STATE is "good"
                    //     If it is, we can quit early!
                    if ( this->_game->isGoodStateToBeIn(nextInLine) ) {

                        // Create int for the total score of this STATE
                        int stateScore = 1;

                        // If so, update solution map with best choice...
                        _solution[state] = new CHOICE( choices[i] );

                        // Since this will be the only choice we make,
                        //     Check to see if that state has been cached...

                        // Iterate the cache to check to see if we have looked
                        //     at this STATE yet.
                        if ( _cache.find( nextInLine ) != _cache.end() ) {
                            // If we have, get the score!
                            stateScore += _cache[ nextInLine ];

                        } else {
                            // Iterate the solution based on that CHOICE...
                            stateScore += iterateSolution(
                                    nextInLine, false );

                            // Don't forget to cache your findings!
                            _cache[ nextInLine ] = stateScore;
                        }

                        // Then finally return true to signify
                        //     that you have found a good path.
                        return stateScore;

                    } else {
                        // Else, add the next STATE to the vector
                        nextStates.push_back(nextInLine);
                    }
                }

                // Since there were no instant choices...

                // Create an int for the best score out of all the CHOICEs
                int bestScore;

                // Create switch for determining if a best score exists:
                bool existingBestScore = false;

                // Create best CHOICE variable
                CHOICE* bestChoice = new CHOICE();

                // Loop through all of the next states
                for (unsigned int i = 0; i < nextStates.size(); i++) {

                    // Create an int for the score of the CHOICE
                    int choiceScore = 0;

                    // Iterate the cache to check to see if we have looked
                    //     at this STATE yet.
                    if ( _cache.find( nextStates[i] ) != _cache.end() ) {
                        // If we have, get the score!
                        choiceScore = _cache[ nextStates[i] ];

                    } else {
                        // If the STATE has not been cached, do so...

                        // Iterate the solution based on that CHOICE...
                        choiceScore += iterateSolution(
                                nextStates[i], false );

                        // Don't forget to cache your findings!
                        _cache[ nextStates[i] ] = choiceScore;
                    }

                    // Test if the CHOICE was better than the best score
                    if ( i == 0 ) {
                        // For the first CHOICE, set it to the best CHOICE
                        bestChoice = new CHOICE( choices[i] );
                        bestScore = choiceScore;

                    } else if ( choiceScore > bestScore ) {
                        // If this CHOICE is better than the last:
                        if ( bestChoice ) {
                            // Deallocate the old pointer first...
                            delete bestChoice;
                        }
                        // Update the bestChoice pointer.
                        bestChoice = new CHOICE( choices[i] );
                        bestScore = choiceScore;
                        // Switch our boolean to let us know that
                        //     the CHOICEs weren't all the same.
                        existingBestScore = true;

                    } else if ( choiceScore < bestScore ) {

                        // Switch our boolean to let us know that
                        //     the CHOICEs weren't all the same.
                        existingBestScore = true;
                    }
                }

                // Update solution map with best choice...
                if ( existingBestScore == false ) {
                    // If there was no score that was better than the rest,
                    //     return the default CHOICE to signify no best
                    //     CHOICE:
                    if ( bestChoice ) {
                        // Deallocate the old pointer first...
                        delete bestChoice;
                    }
                    _solution[ state ] = new CHOICE();

                } else {
                    // There was a CHOICE better than the rest...
                    _solution[ state ] = bestChoice;
                }

                // Return the bestScore for the STATE
                return bestScore;

            } else {
                // Else it is their turn...

                // Create a holder for the total score of the STATE
                int bestScore = -1000;

                // Iterate through the choices set to get a
                //     vector of next STATEs
                for (unsigned int i = 0; i < choices.size(); i++) {

                    // Get the STATE based on that CHOICE
                    STATE nextState = this->_game->getState(
                            state, choices[i] );

                    // int holder for the score of the CHOICE made
                    int choiceScore( this->_game->isGoodStateToBeIn(
                            nextState ) * -1 );

                    // Check to see if we have a solution for that state.
                    CHOICE* choice = _solution[nextState];

                    // If it hasn't been looked at yet, call iterate!
                    if ( ! choice ) {
                        // Iterate through the solution at the next state
                        //    while testing if there are any good paths.
                        choiceScore += iterateSolution( nextState, true );
                    }

                    if ( choiceScore > bestScore ) {
                        bestScore = choiceScore;
                    }
                }

                // Return whether or not you hit a good path.
                return bestScore;
            }
        } /* Computer : iterateSolution */

        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////
    private: /* Computer : Instance variables */
        //////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////

        std::map<STATE, int> _cache; // Cache for memoization

        std::map<STATE, CHOICE*> _solution;
                                // The possible states and the
                                // best choice for those states.

    }; /* Computer class */

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
private: /* Game : Private methods */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

    /*
     * updateState will update the current STATE of the Game based
     *     on the provided CHOICE.
     *
     * @precondition: choice is a valid move
     *
     * @param choice the move that the player made
     */
    void updateState(const CHOICE& choice) {

        // Call setState with the updated STATE
        setState( getState( currentState(), choice ) );

    } /* Game : updateState */

    /*
     * resetState will update the current STATE of the Game based
     *     on the provided STATE.
     *
     * @param choice the move that the player made
     */
    void setState(const STATE& state) {

        // Check for NULL pointers:
        if ( _state ) {

            // Delete the current STATE pointer.
            delete _state;
        }
        _state = new STATE( state );
    } /* Game : resetState */

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////
private: /* Game : Instance variables */
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

    STATE* _state; // pointer to current STATE of Game.

}; /* Game class */

#endif /* GAME_H_ */
